/******************************************************************************
 *
 * Mass-conserving ProblemStat for AMDiS
 *
 * Copyright (C) 2013 Dresden University of Technology. All Rights Reserved.
 * Web: https://fusionforge.zih.tu-dresden.de/projects/amdis/trunk/extensions
 *
 * Author: Simon Praetorius
 *
 * This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
 * WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 *
 * See also license.opensource.txt in the distribution.
 * 
 ******************************************************************************/

/** \file ProblemStatMassConserve2.h */


#ifndef PROBLEM_STAT_MASS_CONSERVE
#define PROBLEM_STAT_MASS_CONSERVE

#include "ExtendedProblemStat.h"
#include "Refinement.h"

namespace AMDiS {

/// Implementation of ProblemStat to allow for the conservation of mass of one
/// solution component, i.e. project the old-solution to the new mesh so that
/// the scalar-product is conserved: \f$ (u^*, v) = (u^\text{old}, v) \f$, where \f$ u^*,v \f$ live on the
/// new mesh and \f$ u^\text{old} \f$ lives on the old mesh.
struct ProblemStatMassConserve2 : public ExtendedProblemStat, public StandardRefineOperation
{
  /// constructor: create a temporary problem that is initialized with 
  /// similar parameters as the regular one
  ProblemStatMassConserve2(std::string name_)
  : ExtendedProblemStat(name_),
    prob2(NULL),
    comp(0)
  {    
    Parameters::get(name_ + "->mass conservation component", comp);
    
    // set parameters for the temporary problem
    std::string name2 = name_ + "_tmp";
    int zero = 0;
    int one = 1;
    int two = 2;
    double tol = 1.e-8;
    int degree = 1;
    int dim = 1;
    std::string meshName = "";
    
    Parameters::get(name_ + "->polynomial degree[" + boost::lexical_cast<std::string>(comp) + "]", degree);
    Parameters::get(name_ + "->dim", dim);
    Parameters::get(name_ + "->mesh", meshName);
    
    Parameters::set(name2 + "->components", one);
    Parameters::set(name2 + "->polynomial degree[0]", degree);
    Parameters::set(name2 + "->dim", dim);
    Parameters::set(name2 + "->mesh", meshName);
#if defined HAVE_UMFPACK || defined HAVE_PARALLEL_PETSC || defined HAVE_PETSC
    std::string solverName = "direct";
#else
    std::string solverName = "cg";
#endif
    Parameters::set(name2 + "->solver", solverName);
    Parameters::set(name2 + "->tolerance", tol);
    Parameters::set(name2 + "->info", zero);
    
    // create a new temporary problem
    prob2 = new ProblemStat(name2);
  }
  
  
  ~ProblemStatMassConserve2()
  {
    if (prob2 != NULL) {
      delete prob2;
      prob2 = NULL;
    }
  }
  
  
  /// make shure that temporary problem has an equal mesh as regular problem
  /// and initialize this mesh only once
  void initialize(Flag initFlag,
		  ProblemStatSeq *adoptProblem = NULL,
		  Flag adoptFlag = INIT_NOTHING) override
  {
    ExtendedProblemStat::initialize(initFlag - INIT_MESH, adoptProblem, adoptFlag); // create meshes for prob
    
    std::vector<Mesh*> meshes2;
    meshes2.push_back(getMesh(comp));
    prob2->setComponentMesh(0, getMesh(comp));
    prob2->setMeshes(meshes2);    
    prob2->initialize(INIT_ALL - INIT_MESH);
    
    for (size_t i = 0; i < meshes2.size(); i++) {
      int globalRefinements = 0;

      // If AMDiS is compiled for parallel computations, the global refinements are
      // ignored here. Later, each rank will add the global refinements to its
      // private mesh.
#ifndef HAVE_PARALLEL_DOMAIN_AMDIS
      Parameters::get(meshes2[i]->getName() + "->global refinements",
		      globalRefinements);
#endif

      bool initMesh = true;

      // Initialize the meshes if there is no serialization file.
      if (initMesh && meshes2[i] && !(meshes2[i]->isInitialized())) {
	meshes2[i]->initialize();
	prob2->getRefinementManager()->globalRefine(meshes2[i], globalRefinements);
      }
    }
    
    prob2->getRhsVector(comp)->setCoarsenOperation(COARSE_RESTRICT);
    
    fillOperators();
  }
  
  /// implementation of StandardRefineOperation::beforeCoarsen
  void beforeCoarsen(AdaptInfo* adaptInfo, Flag markFlag) override
  {
    if (markFlag.isSet(MESH_COARSENED))
      prob2->buildAfterCoarsen(adaptInfo, markFlag, false, true);
  }
  
  /// in the first iteration copy initial-solution to temporary problem
  /// and prepare the temporary mesh
  void buildBeforeCoarsen(AdaptInfo* adaptInfo, Flag markFlag) override
  {   
    ExtendedProblemStat::buildBeforeCoarsen(adaptInfo, markFlag); 
    beforeRefine(adaptInfo, markFlag);
  }
  
  
  /// implementation of StandardRefineOperation::afterCoarsen
  void afterCoarsen(AdaptInfo* adaptInfo, Flag markFlag) override
  {
    if (markFlag.isSet(MESH_COARSENED)) {
      prob2->buildAfterCoarsen(adaptInfo, markFlag, true, false);
      prob2->solve(adaptInfo);
      *getSolution(comp) = *prob2->getSolution(0);
    }
  }
  
  /// solve temporary problem to project solution to new mesh and then
  /// copy temporary solution back to regular solution
  void buildAfterCoarsen(AdaptInfo* adaptInfo, 
			 Flag markFlag, 
			 bool asmM, 
			 bool asmV) override
  {  
    afterCoarsen(adaptInfo, markFlag);
    
    // call original buildAfterCoarsen
    ExtendedProblemStat::buildAfterCoarsen(adaptInfo, markFlag, asmM, asmV); 
  }
  
  /// add mass-matrix and transfer-operator to temporary problem, to
  /// project solution to new mesh
  void fillOperators()
  {
    // conversion propblem 2
    const FiniteElemSpace* feSpace2 = prob2->getFeSpace(0);

    Operator *opMnew0 = new Operator(feSpace2, feSpace2);
    opMnew0->addTerm(new Simple_ZOT);
    Operator *opMold01 = new Operator(feSpace2, feSpace2);
    opMold01->addTerm(new VecAtQP_ZOT(getSolution(comp)));
    
    // assemble only one block of the system
    prob2->addMatrixOperator(*opMnew0,0,0);
    prob2->addVectorOperator(*opMold01, 0); 
  }
  
private:  
  /// temporary problem for the projection of the old-solution to the new mesh
  ProblemStat* prob2;
  
  /// component the should be copnserved
  int comp;
};

}

#endif // PROBLEM_STAT_MASS_CONSERVE