/******************************************************************************
 *
 * Extension of AMDiS - Adaptive multidimensional simulations
 *
 * Copyright (C) 2013 Dresden University of Technology. All Rights Reserved.
 * Web: https://fusionforge.zih.tu-dresden.de/projects/amdis
 *
 * Authors: Simon Praetorius et al.
 *
 * This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
 * WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 *
 * See also license.opensource.txt in the distribution.
 * 
 ******************************************************************************/

#include "Helpers.h"

namespace AMDiS { namespace base_problems {

namespace detail {

template<typename P> 
PolarizationField<P>::PolarizationField(const std::string &name_) :
  super(name_, true),
  vectorField(NULL),
  oldTimestep(0.0),
  minus1(-1.0),
  alpha2(1.0),
  alpha4(1.0),
  C1(0.2),
  C4(0.0),
  epsilon(0.0),
  K(1.0),
  fileWriter(NULL)
{
  oldSolution.resize(self::dow);
  for (size_t i = 0; i < self::dow; i++)
    oldSolution[i] = NULL;
  
  Parameters::get(self::name + "->alpha2", alpha2);
  Parameters::get(self::name + "->alpha4", alpha4);
  
  Parameters::get(self::name + "->C1", C1);
  Parameters::get(self::name + "->C4", C4);
  
  Parameters::get(self::name + "->epsilon", epsilon);
  if (epsilon > 0.0) {
    epsInv = 1.0/epsilon;
    C1 = -epsInv;
    C4 = epsInv;
  }
  
  
  Parameters::get(self::name + "->K", K);
}

template<typename P> 
PolarizationField<P>::~PolarizationField() 
{   
  if (vectorField != NULL) {
    delete vectorField;
    vectorField = NULL;
  }
  
  for (size_t i = 0; i < oldSolution.size(); i++) {
    if (oldSolution[i] != NULL)
      delete oldSolution[i];
    oldSolution[i] = NULL;
  }
 
  if (fileWriter) { 
    delete fileWriter;
    fileWriter = NULL;
  }
}


template<typename P> 
void PolarizationField<P>::initData()
{ 
  if (vectorField == NULL)
    vectorField = new DOFVector<WorldVector<double> >(self::getFeSpace(0), "vectorField");
  
  for (size_t i = 0; i < self::dow; i++)
    oldSolution[i] = new DOFVector<double>(self::getFeSpace(i), "old(v_"+Helpers::toString(i)+")");
  
#ifdef HAVE_PARALLEL_DOMAIN_AMDIS
  for (size_t i = 0; i < self::dow; i++)
    Parallel::MeshDistributor::globalMeshDistributor->addInterchangeVector(oldSolution[i]);
#endif
    
  fileWriter = new FileVectorWriter(self::name + "->vectorField->output", self::getFeSpace()->getMesh(), vectorField);

  super::initData();
}


template<typename P> 
void PolarizationField<P>::transferInitialSolution(AdaptInfo *adaptInfo)
{ 
  calcVectorField();
  for (size_t i = 0; i < self::dow; i++)
    oldSolution[i]->copy(*self::prob->getSolution()->getDOFVector(i));
  
  super::transferInitialSolution(adaptInfo);
}


template<typename P> 
void PolarizationField<P>::fillOperators()
{ 
  WorldVector<DOFVector<double>* > vec;
  for (size_t k = 0; k < self::dow; k++)
    vec[k] = self::prob->getSolution()->getDOFVector(k);
  
  const FiniteElemSpace* feSpace = self::getFeSpace(0);
  
  // fill operators for component P
  for (size_t i = 0; i < self::dow; ++i) {
    /// < (1/tau)*P_i , psi >
    Operator *opTime = new Operator(feSpace, feSpace);
    addZOT(opTime, 1.0);
    self::prob->addMatrixOperator(*opTime, i, i, self::getInvTau(), self::getInvTau());
    
    /// < (1/tau)*P_i^old , psi >
    Operator *opTimeOld = new Operator(feSpace, feSpace);
    addZOT(opTimeOld, valueOf(getOldSolution(i)));
    self::prob->addVectorOperator(*opTimeOld, i, self::getInvTau(), self::getInvTau());

    /// Diffusion-Operator
    Operator *opLaplace = new Operator(feSpace, feSpace);
    addSOT(opLaplace, alpha2);
    addZOT(opLaplace, alpha4);
    self::prob->addMatrixOperator(*opLaplace, i, i+self::dow);
  }

  // fill operators for component P#
  for (size_t i = 0; i < self::dow; ++i) {
    /// < P# , psi >
    Operator *opM = new Operator(feSpace, feSpace);
    addZOT(opM, 1.0);
    self::prob->addMatrixOperator(*opM, i+self::dow, i+self::dow);
    
    /// < (-C1 - C4*P^2)*P , psi >
    Operator *opNonlin = new Operator(feSpace, feSpace);
    addZOT(opNonlin, (-C1 - C4 * unary_dot(valueOf(vectorField))) );
    self::prob->addMatrixOperator(*opNonlin, i+self::dow, i);
    
    for (size_t j = 0; j < self::dow; ++j) {
      Operator *opNonlin2 = new Operator(feSpace, feSpace);
      addZOT(opNonlin2, (-2.0*C4) * valueOf(vec[i]) * valueOf(vec[j]));
      self::prob->addMatrixOperator(*opNonlin2, i+self::dow, j);
    }
    
    Operator *opNonlin3 = new Operator(feSpace, feSpace);
    addZOT(opNonlin3, (-2.0*C4) * valueOf(vec[i]) * unary_dot(valueOf(vectorField)) );
    self::prob->addVectorOperator(*opNonlin3, i+self::dow);
  }
  
  fillLaplacian();

  Operator *opNull = new Operator(feSpace, feSpace);
  addZOT(opNull, 0.0);
  
  for (size_t i = 0; i < self::dow; ++i) {
    for (size_t j = i+1; j < self::dow; ++j) {
      self::prob->addMatrixOperator(*opNull, i, j);
      self::prob->addMatrixOperator(*opNull, j, i);
      self::prob->addMatrixOperator(*opNull, i, j+self::dow);
      self::prob->addMatrixOperator(*opNull, j, i+self::dow);
      self::prob->addMatrixOperator(*opNull, i+self::dow, j+self::dow);
      self::prob->addMatrixOperator(*opNull, j+self::dow, i+self::dow);
    }
  }
}


template<typename P> 
void PolarizationField<P>::fillLaplacian()
{
  const FiniteElemSpace* feSpace = self::getFeSpace(0);
  for (size_t i = 0; i < self::dow; ++i) {
    /// < -K*grad(P) , grad(psi) >
    Operator *opL = new Operator(feSpace, feSpace);
    addSOT(opL, -K);
    self::prob->addMatrixOperator(*opL, i+self::dow, i);
  }
}



template<typename P> 
void PolarizationField<P>::closeTimestep(AdaptInfo *adaptInfo)
{
  calcVectorField();
  for (size_t i = 0; i < self::dow; i++)
    oldSolution[i]->copy(*self::prob->getSolution()->getDOFVector(i));
  
  super::closeTimestep(adaptInfo);
}


template<typename P> 
void PolarizationField<P>::writeFiles(AdaptInfo *adaptInfo, bool force)
{
  super::writeFiles(adaptInfo, force);
  self::fileWriter->writeFiles(adaptInfo, false);
}

} // end namespace detail

} }
