
#include "GenericOperatorTerm.h"

namespace AMDiS { namespace base_problems {
  
  namespace detail {
  
    using namespace std;

    template<typename P>
    QuasiCrystal<P>::QuasiCrystal(const std::string &name_) :
      super(name_),
      q(2.0*std::cos(m_pi/12.0)),
      r(-0.015),
      c(50.0),
      alpha(1.0),
      density(0.0),	// mean density
      two(2.0),
      minus2(-2.0),
      three(3.0),
      minus3(-3.0)
    {  
      Parameters::get(name_ + "->r",r);
      Parameters::get(name_ + "->c", c);
      Parameters::get(name_ + "->q", q);
      Parameters::get(name_ + "->density", density);
      Parameters::get(name_ + "->alpha", alpha);
    }


    template<typename P>
    void QuasiCrystal<P>::fillOperators()
    {      
      const FiniteElemSpace* feSpace = self::getFeSpace(0);
      
      DOFVector<double>* rho = self::prob->getSolution()->getDOFVector(0);
//       DOFVector<double>* mu = self::prob->getSolution()->getDOFVector(1);
      
      // mass-matrix
      Operator *opM = new Operator(feSpace, feSpace);
      addZOT(opM, 1.0);

      // laplace-matrix
      Operator *opL = new Operator(feSpace, feSpace);
      addSOT(opL, 1.0);
      
      // mu - c*laplace(nu) - laplace1(rho) + lin(rho) + [2rho'-3(rho')^2]*rho = (rho')^2-2*(rho')^3
      // --------------------------------------------------------------------------------------------
      Operator *op00 = new Operator(feSpace, feSpace);
      addSOT(op00, (2.0*sqr(q)*c*(1.0+sqr(q))));
      addZOT(op00, (-c*sqr(sqr(q)) - r) + 2.0 * valueOf(rho) - 3.0 * pow<2>(valueOf(rho)));      
      self::prob->addMatrixOperator(op00, 0, 0); 
      
      self::prob->addMatrixOperator(opM, 0, 1); // mu
      self::prob->addMatrixOperator(opL, 0, 2, &c, &c); // -c*laplace(nu)
      
      Operator *op0 = new Operator(feSpace);
      addZOT(op0, pow<2>(valueOf(rho)) - 2.0 * pow<3>(valueOf(rho)));
      self::prob->addVectorOperator(op0, 0);
      
      // dt(rho) - laplace(mu) = 0
      // -----------------------------------
      self::prob->addMatrixOperator(opM, 1, 0, self::getInvTau(), self::getInvTau());
      self::prob->addMatrixOperator(opL, 1, 1); // -laplace(mu)
      
      // . . . vectorOperators . . . . . . . . . . . . . . .
      Operator *opMold = new Operator(feSpace);
      addZOT(opMold, valueOf(rho));
      self::prob->addVectorOperator(opMold, 1, self::getInvTau(), self::getInvTau());
      
      // nu - laplace2(rho) - laplace(psi) = 0
      // ----------------------------------------------------------------------
      Operator *opL2 = new Operator(feSpace, feSpace);
      addSOT(opL2, (sqr(1.0+sqr(q))+2.0*sqr(q)));
      self::prob->addMatrixOperator(opL2, 2, 0); // -laplace2(rho)
      
      self::prob->addMatrixOperator(opM, 2, 2); // nu
      self::prob->addMatrixOperator(opL, 2, 3); // -laplace(psi)
      
      // psi - laplace3(rho) - laplace(theta) = 0
      // ----------------------------------------------------------------------
      Operator *opL3 = new Operator(feSpace, feSpace);
      addSOT(opL3, (2.0*(1.0+sqr(q))));
      self::prob->addMatrixOperator(opL3, 3, 0); // -laplace3(rho)
      
      self::prob->addMatrixOperator(opM, 3, 3); // psi
      self::prob->addMatrixOperator(opL, 3, 4); // -laplace(theta)
      
      // theta - laplace(rho) = 0
      // ----------------------------------------------------------------------
      self::prob->addMatrixOperator(opL, 4, 0); // -laplace(rho)
      self::prob->addMatrixOperator(opM, 4, 4); // theta
    }

    
    template<typename P>
    void QuasiCrystal<P>::finalizeData()
    {
      self::setAssembleMatrixOnlyOnce_butTimestepChange(0,1);
      self::setAssembleMatrixOnlyOnce_butTimestepChange(0,2);
      
      self::setAssembleMatrixOnlyOnce_butTimestepChange(1,0);
      self::setAssembleMatrixOnlyOnce_butTimestepChange(1,1);
      
      self::setAssembleMatrixOnlyOnce_butTimestepChange(2,0);
      self::setAssembleMatrixOnlyOnce_butTimestepChange(2,2);
      self::setAssembleMatrixOnlyOnce_butTimestepChange(2,3);
      
      self::setAssembleMatrixOnlyOnce_butTimestepChange(3,0);
      self::setAssembleMatrixOnlyOnce_butTimestepChange(3,3);
      self::setAssembleMatrixOnlyOnce_butTimestepChange(3,4);
      
      self::setAssembleMatrixOnlyOnce_butTimestepChange(4,0);
      self::setAssembleMatrixOnlyOnce_butTimestepChange(4,4);
    }
    
  } // end namespace detail

} } // end namespaces
