/******************************************************************************
 *
 * Extension of AMDiS - Adaptive multidimensional simulations
 *
 * Copyright (C) 2013 Dresden University of Technology. All Rights Reserved.
 * Web: https://fusionforge.zih.tu-dresden.de/projects/amdis
 *
 * Authors: Simon Praetorius et al.
 *
 * This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
 * WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 *
 * See also license.opensource.txt in the distribution.
 * 
 ******************************************************************************/

#include "AMDiS.h"
#include "CahnHilliard_DD.h"
#include "Expressions.h"

#ifdef HAVE_PARALLEL_DOMAIN_AMDIS
#include "parallel/PetscSolverCahnHilliard.h"
#endif

using namespace AMDiS;

class MyCahnHilliard : public CahnHilliard_DD
{
  typedef CahnHilliard_DD super;
public:
  MyCahnHilliard(std::string name)
  : super(name), phaseField(NULL), fileWriterPhase(NULL), filename(""), delta(1.e-5)
  {
    Parameters::get(name + "->phasefield->filename", filename);
    Parameters::get(name + "->phasefield->perturbation", delta);
  }
  
  ~MyCahnHilliard() 
  {
    if (phaseField) {
      delete phaseField;
      phaseField = NULL;
    }
    if (fileWriterPhase) {
      delete fileWriterPhase;
      fileWriterPhase = NULL;
    }
  }
  
  void initData() override
  {
    super::initData();
    
    phaseField = new DOFVector<double>(getFeSpace(0), "phaseField"); 
    super::setPhase(phaseField);
    
    fileWriterPhase = new FileWriter(name + "->phasefield->output", getFeSpace(0)->getMesh(), phaseField);

  }
  
  void finalizeData() override
  {
#ifdef HAVE_PARALLEL_DOMAIN_AMDIS
    Parallel::PetscSolverCahnHilliard* solver = dynamic_cast<Parallel::PetscSolverCahnHilliard*>(prob->getSolver());
    if (solver) {
      solver->setChData2(&eps, &gamma, getTau(), phaseField);
    }
#endif
      
    super::finalizeData();
  }
  
  void solveInitialProblem(AdaptInfo *adaptInfo) override
  {
    io::Arh2Reader::readFile(filename, *phaseField, false);
#ifdef HAVE_PARALLEL_DOMAIN_AMDIS
    Parallel::MeshDistributor::globalMeshDistributor->checkMeshChange();
#endif
    
    io::writeFile(phaseField, "phaseField.vtu");
//     phaseField->set(1.0);
    
    // random initial data
    prob->getSolution(1)->interpol(new AMDiS::Random(0.0, 1.0));
    
    *phaseField << max(valueOf(phaseField), delta);
  }  


  void writeFiles(AdaptInfo *adaptInfo, bool force)
  { 
    super::writeFiles(adaptInfo, force);
    fileWriterPhase->writeFiles(adaptInfo, false);
  }
  
protected:  
  DOFVector<double>* phaseField;
  FileWriter* fileWriterPhase;
  std::string filename;
  double delta;
};



int main(int argc, char** argv)
{ FUNCNAME("main");

  AMDiS::init(argc, argv);   
  Timer t;
  
#ifdef HAVE_PARALLEL_DOMAIN_AMDIS
  CreatorMap< LinearSolverInterface >::addCreator("p_petsc_petsc-ch", new Parallel::PetscSolverCahnHilliard::Creator);
#endif
  
  MyCahnHilliard chProb("ch");
  chProb.initialize(INIT_ALL);
  chProb.initTimeInterface();
  
  // Adapt-Infos
  AdaptInfo adaptInfo("adapt", chProb.getNumComponents());
  AdaptInstationary adaptInstat("adapt", chProb, adaptInfo, chProb, adaptInfo);

  int error_code = adaptInstat.adapt(); 

  MSG("elapsed time= %d sec\n", t.elapsed());
  AMDiS::finalize();

  return error_code;
}
