#include "AMDiS.h"

inline double pow2(const double &v) 
{
  return v * v;
}

inline double pow4(const double &v) 
{
  return v * v * v * v;
}

// === Anisotropy functions for 2D ===

class Aniso2d_Square : public AbstractFunction<double, WorldVector<double> >
{
public:
  Aniso2d_Square(double g)
    : AbstractFunction<double, WorldVector<double> >(2),
      gamma(g),
      anisoReg(1e-6)
  {
    eps2 = 1.0 - 3.0 * gamma;
    eps4 = 4.0 * gamma / (1.0 - 3.0 * gamma);
  }

  double operator()(const WorldVector<double>& n) const
  {
    double eta = eps2 * 
      (1.0 + eps4 * (pow4(n[0]) + pow4(n[1])) / (pow4(norm(&n)) + anisoReg));

    return eta * eta;
  }

private:
  double gamma;

  double anisoReg;

  double eps2, eps4;
};


class Aniso2d_Laplace : public AbstractFunction<WorldMatrix<double>, WorldVector<double> >
{
public:
  Aniso2d_Laplace(double g) 
    : AbstractFunction<WorldMatrix<double>, WorldVector<double> >(2),
      gamma(g),
      anisoReg(1e-6)
  {
    eps2 = 1.0 - 3.0 * gamma;
    eps4 = 4.0 * gamma / (1.0 - 3.0 * gamma);    
  }
  
  WorldMatrix<double> operator()(const WorldVector<double>& n) const 
  {
    WorldMatrix<double> result;
    result.set(0.0);

    double d = (pow4(n[0]) + pow4(n[1])) / (pow4(norm(&n)) + anisoReg);
    double normSquare = pow2(norm(&n)) + anisoReg;
    double eta = eps2 * (1 + eps4 * d);

    result[0][0] = 
      eta * (eta + 16.0 * gamma * (pow2(n[0]) / normSquare - d));

    result[1][1] = 
      eta * (eta + 16.0 * gamma * (pow2(n[1]) / normSquare - d));

    return result;
  }    

private:
  double gamma;

  double anisoReg;

  double eps2, eps4; 
};


// === Anisotropy functions for 3D ===

class Aniso3d_Square : public AbstractFunction<double, WorldVector<double> >
{
public:
  Aniso3d_Square(double g)
    : AbstractFunction<double, WorldVector<double> >(2),
      gamma(g),
      anisoReg(1e-6)
  {
    eps2 = 1.0 - 3.0 * gamma;
    eps4 = 4.0 * gamma / (1.0 - 3.0 * gamma);
  }

  double operator()(const WorldVector<double>& n) const
  {
    double eta = eps2 * 
      (1.0 + eps4 * (pow4(n[0]) + pow4(n[1]) + pow4(n[2])) / (pow4(norm(&n)) + anisoReg));

    return eta * eta;
  }

private:
  double gamma;

  double anisoReg;

  double eps2, eps4;
};


class Aniso3d_Laplace : public AbstractFunction<WorldMatrix<double>, WorldVector<double> >
{
public:
  Aniso3d_Laplace(double g) 
    : AbstractFunction<WorldMatrix<double>, WorldVector<double> >(2),
      gamma(g),
      anisoReg(1e-6)
  {
    eps2 = 1.0 - 3.0 * gamma;
    eps4 = 4.0 * gamma / (1.0 - 3.0 * gamma);    
  }
  
  WorldMatrix<double> operator()(const WorldVector<double>& n) const 
  {
    WorldMatrix<double> &result = const_cast<WorldMatrix<double>&>(tmpWorldMat);
    result.set(0.0);

    double d = (pow4(n[0]) + pow4(n[1]) + pow4(n[2])) / (pow4(norm(&n)) + anisoReg);
    double normSquare = pow2(norm(&n)) + anisoReg;
    double eta = eps2 * (1 + eps4 * d);

    result[0][0] = 
      eta * (eta + 16.0 * gamma * (pow2(n[0]) / normSquare - d));

    result[1][1] = 
      eta * (eta + 16.0 * gamma * (pow2(n[1]) / normSquare - d));

    result[2][2] = 
      eta * (eta + 16.0 * gamma * (pow2(n[2]) / normSquare - d));

    return result;
  }    

private:
  WorldMatrix<double> tmpWorldMat;

  double gamma;

  double anisoReg;

  double eps2, eps4; 
};


class NullFct : public AbstractFunction<WorldVector<double>, WorldMatrix<double> >
{
public:
    
  NullFct()
    : AbstractFunction<WorldVector<double>, WorldMatrix<double> >(1)
  {
    result.set(0.0);
  }

  WorldVector<double> operator()(const WorldMatrix<double>& n) const 
  {   
    return result;
  }

private:
  WorldVector<double> result;
};
