/******************************************************************************
 *
 * Extension of AMDiS - Adaptive multidimensional simulations
 *
 * Copyright (C) 2013 Dresden University of Technology. All Rights Reserved.
 * Web: https://fusionforge.zih.tu-dresden.de/projects/amdis
 *
 * Authors: Simon Praetorius et al.
 *
 * This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
 * WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 *
 * See also license.opensource.txt in the distribution.
 * 
 ******************************************************************************/
#include "AMDiS.h"
#include "NavierStokesPhase_TaylorHood.h"
#include "LinearElasticity.h"

#include "ElasticityNavierStokes.h"
#include "navierStokes.h"
// #include "time/ExtendedRosenbrockAdaptInstationary.h"
#include "Refinement.h"
#include "MeshFunction_Level.h"
#include "PhaseFieldConvert.h"
#include "kdtree_nanoflann_dof.h"
#include "Views.h"

#include "boost/date_time/posix_time/posix_time.hpp"

using namespace AMDiS;
using namespace boost::posix_time;

struct NormalStress : TertiaryAbstractFunction<double, WorldVector<double>, WorldVector<double>, WorldVector<double> >
{
  NormalStress(int comp_, double viscosity_) : comp(comp_), viscosity(viscosity_) {}

  double operator()(const WorldVector<double>& normal, const WorldVector<double>& grdU0, const WorldVector<double>& grdU1) const
  {
	return viscosity * (2.0*grdU0[comp]*normal[comp] + (grdU1[0] + grdU0[1])*normal[1-comp]);
  }
private:
  int comp;
  double viscosity;
};

struct NormalStressP : TertiaryAbstractFunction<double, WorldVector<double>, double, double>
{
  NormalStressP(int comp_) : comp(comp_) {}

  double operator()(const WorldVector<double>& normal, const double& stress, const double& p) const
  {
    return stress - p*normal[comp];
  }
private:
  int comp;
};


struct VectorBC : AbstractFunction<double, WorldVector<double> >
{
  VectorBC(int comp_, DOFVector<WorldVector<double> >* velocity_)
  : AbstractFunction<double, WorldVector<double> >(4),
    comp(comp_),
    velocity(velocity_) {
	  TEST_EXIT(velocity != NULL)("pointer auf DOFVector<WorldVector> ist NULL\n");
    }

  double operator()(const WorldVector<double>& x) const
  {
	WorldVector<double> p = x;
    WorldVector<double> localVel = evalAtPoint<WorldVector<double> >(*velocity, p);
    return localVel[comp];
  }
private:
  int comp;
  DOFVector<WorldVector<double> >* velocity;
};

///_______________________________________________________________________________________________________________________

class Elasticity_Obstacle : public LinearElasticity
{
public:
  typedef LinearElasticity super;

public:
  Elasticity_Obstacle(std::string name_) : super(name_) {}

  ~Elasticity_Obstacle()
  {
	for (size_t i = 0; i < dow; i++)
	  delete stress[i];
  }

  void initData()
  {
    super::initData();
    dim = getMesh()->getDim();

	for (size_t i = 0; i < dow; i++) {
	  stress[i] = new DOFVector<double>(getFeSpace(0), "stress_"+Helpers::toString(i));
	  stress[i]->set(0.0);
	}
    for (size_t i = 0; i < dow; i++) {
	  deformation[i] = prob->getSolution()->getDOFVector(i);
	  deformationVelocity[i] = prob->getSolution()->getDOFVector(dow+i);
	}
	
	std::vector<DOFVector<double>*> vecs;
	for (size_t i = 0; i < dow; i++)
	  vecs.push_back(stress[i]);
	fileWriterStress = new FileWriter(name + "->stress->output", getFeSpace()->getMesh(), vecs);
  }

  DOFVector<double>* getDeformation(int i)
  {
    return deformation[i];
  }
  WorldVector<DOFVector<double>*> getDeformation()
  {
	return deformation;
  }

  DOFVector<double>* getDeformationVelocity(int i)
  {
    return deformationVelocity[i];
  }
  WorldVector<DOFVector<double>*> getDeformationVelocity()
  {
	return deformationVelocity;
  }

  void setNormalStress(DOFVector<WorldVector<double> >* stress_)
  {
    boundaryStress = stress_;
  }
  
  void transferInitialSolution(AdaptInfo* adaptInfo)
  {
    super::transferInitialSolution(adaptInfo);
  }

  void initTimestep(AdaptInfo* adaptInfo)
  {
    super::initTimestep(adaptInfo);
	TEST_EXIT(boundaryStress != NULL)("NULL-Pointer problem!\n");
	for (size_t i = 0; i < dow; i++)
	  transformDOF_coords(*boundaryStress, *stress[i], new AMDiS::Component2<>(i));
	fileWriterStress->writeFiles(adaptInfo, false);
  }
  
  void closeTimestep(AdaptInfo* adaptInfo)
  {
    super::closeTimestep(adaptInfo);
  }

  void fillBoundaryConditions()
  {
//     super::fillBoundaryConditions();

    AbstractFunction<double, WorldVector<double> > *zero = new AMDiS::Const<double, WorldVector<double> >(0.0);
    double zeroValue = 0.0;

    //  +--+
    //  |  |   1
    //10|  +==========
    //  |  | 
    //  +--+
    
    for (int i = 0; i < dow; i++) {
      prob->addDirichletBC(10, i, i, zero);
      prob->addNeumannBC(1, i, i, stress[i]);
    }
  }

private:
  WorldVector<DOFVector<double>*> deformation;
  WorldVector<DOFVector<double>*> deformationVelocity;
  WorldVector<DOFVector<double>*> stress;
  DOFVector<WorldVector<double> >* boundaryStress;
  
  FileWriter* fileWriterStress;
};

///_______________________________________________________________________________________________________________________

class Elasticity_Fluid : public StandardBaseProblem
{
public:
  typedef StandardBaseProblem super;

public:
  Elasticity_Fluid(std::string name_) : super(name_) {}

  ~Elasticity_Fluid()
  {
	for (size_t i = 0; i < dow; i++)
	  delete deformationVec[i];
  }

  void initData()
  {
    super::initData();
    dim = getMesh()->getDim();

	for (size_t i = 0; i < dow; i++) {
	  deformationVec[i] = new DOFVector<double>(getFeSpace(i), "deformation");
	  deformationVec[i]->set(0.0);
	}
	for (size_t i = 0; i < dow; i++)
	  deformation = prob->getSolution()->getDOFVector(i);
  }

  DOFVector<double>* getDeformation(int i)
  {
    return deformation[i];
  }
  WorldVector<DOFVector<double>*> getDeformation()
  {
    return deformation;
  }

  void setDeformation(WorldVector<DOFVector<double>*> deformation_)
  {
    boundaryDeformation = deformation_;
  }
  
  void transferInitialSolution(AdaptInfo* adaptInfo)
  {
    super::transferInitialSolution(adaptInfo);
  }

  void initTimestep(AdaptInfo* adaptInfo)
  {
    super::initTimestep(adaptInfo);
	TEST_EXIT(boundaryDeformation[0] != NULL)("NULL-Pointer problem!\n");
	for (size_t i = 0; i < dow; i++)
	  interpol_coords(*boundaryDeformation[i], *deformationVec[i]);
  }
  
  void closeTimestep(AdaptInfo* adaptInfo)
  {
    super::closeTimestep(adaptInfo);
  }
  
  void fillOperators()
  {
    // ===== create matrix operator =====
    for (size_t i = 0; i < dow; i++) {
	  Operator *matrixOperator = new Operator(getFeSpace());
	  matrixOperator->addTerm(new Simple_SOT);
	  prob->addMatrixOperator(matrixOperator, i, i);
	}
  }

  void fillBoundaryConditions()
  {
    super::fillBoundaryConditions();

    AbstractFunction<double, WorldVector<double> > *zero = new AMDiS::Const<double, WorldVector<double> >(0.0);
    double zeroValue = 0.0;

    //  +--+
    //  |  |   1
    //10|  +==========
    //  |  |
    //  +--+

    for (int i = 0; i < dow; i++) {
      prob->addDirichletBC(10, i, i, zero);
      prob->addDirichletBC(2, i, i, zero);
      prob->addDirichletBC(3, i, i, zero);
      prob->addDirichletBC(4, i, i, zero);
      prob->addDirichletBC(5, i, i, zero);
	  
      prob->addDirichletBC(1, i, i, deformationVec[i]);
    }
  }

private:
  WorldVector<DOFVector<double>*> deformation; // solution
  WorldVector<DOFVector<double>*> boundaryDeformation;
  WorldVector<DOFVector<double>*> deformationVec; // boundaryDeformation->deformationVec
};

///_______________________________________________________________________________________________________________________

class NS_Channel : public NavierStokes_TaylorHood
{
public:
  typedef NavierStokes_TaylorHood super;
  
public:
  NS_Channel(std::string name_) : super(name_) {}
  
  ~NS_Channel()
  {
    delete inflowBC;
	delete fileWriterStress;
	delete stress;
	for (size_t i = 0; i < dow; i++)
	  delete velocityVec[i];
  }

  void initData()
  {
    super::initData();
	stress = new DOFVector<WorldVector<double> >(getFeSpace(0), "stress");
	WorldVector<double> zeroVec; zeroVec.set(0.0);
	stress->set(zeroVec);
	for (size_t i = 0; i < dow; i++) {
	  velocityVec[i] = new DOFVector<double>(getFeSpace(i), "velocity_i");
	  velocityVec[i]->set(0.0);
	}
	  
    dim = getMesh()->getDim();
	fileWriterStress = new FileVectorWriter(name + "->stress->output", getFeSpace()->getMesh(), stress);
  }

  void solveInitialProblem(AdaptInfo* adaptInfo)
  {
    super::solveInitialProblem(adaptInfo);

    inflowBC->setTimePtr(adaptInfo->getTimePtr());
  }
  
  void transferInitialSolution(AdaptInfo* adaptInfo)
  {
	super::transferInitialSolution(adaptInfo);
	fileWriterStress->writeFiles(adaptInfo, false);
  }

  DOFVector<WorldVector<double> >* getNormalStress()
  {
    return stress;
  }

  void setDeformationVelocity(WorldVector<DOFVector<double>*> deformationVel_)
  {
    boundaryVelocity = deformationVel_;
  }
  
  void initTimestep(AdaptInfo* adaptInfo)
  {
    super::initTimestep(adaptInfo);
	TEST_EXIT(boundaryVelocity[0] != NULL)("NULL-Pointer problem!\n");
	for (size_t i = 0; i < dow; i++)
	  interpol_coords(*boundaryVelocity[i], *velocityVec[i]);
  }
  
  void closeTimestep(AdaptInfo* adaptInfo)
  {
    super::closeTimestep(adaptInfo);
    calcNormalStress(1);
	fileWriterStress->writeFiles(adaptInfo, false);
  }

  void fillBoundaryConditions()
  {
    
    // +------ 5 ------+
    // |               |
    // 2   # <--1      3
    // |               |
    // +------ 4 ------+

    AbstractFunction<double, WorldVector<double> > *zero = new AMDiS::Const<double, WorldVector<double> >(0.0);
    

    /// at rigid wall: no-slip boundary condition
    for (size_t i = 0; i < dow; i++) {
      getProblem(0)->addDirichletBC(10, i, i, zero);
      getProblem(0)->addDirichletBC(1, i, i, velocityVec[i]);
      getProblem(0)->addDirichletBC(4, i, i, zero);
      getProblem(0)->addDirichletBC(5, i, i, zero);
    }

    double H = 4.1;
    double Um = 1.5;
    Parameters::get("mesh->H",H);
    Parameters::get("ns->Um",Um);
    inflowBC = new InflowBC(H,Um);

    /// at left wall: prescribed velocity
    getProblem(0)->addDirichletBC(2, 0, 0, inflowBC);
    getProblem(0)->addDirichletBC(2, 1, 1, zero);
	/// at outflow boundary: pressure const 0.0
    getProblem(0)->addDirichletBC(3, 1, 1, zero);
  }

private:
  DOFVector<WorldVector<double> >* stress;
  WorldVector<DOFVector<double>*> boundaryVelocity;
  WorldVector<DOFVector<double>*> velocityVec; // boundaryVelocity->velocityVec
  
  void calcNormalStress(BoundaryType boundaryType = 1)
  {
    const FiniteElemSpace* feSpace = getFeSpace(0);
	Mesh* mesh = feSpace->getMesh();
	
	WorldVector<DOFVector<WorldVector<double> >*> grdU;
	for (size_t i = 0; i < dow; i++) {
	  grdU[i] = new DOFVector<WorldVector<double> >(getFeSpace(i), "grdU_i");
	  getSolution()->getDOFVector(i)->getRecoveryGradient(grdU[i]);
	}
	
    WorldVector<double> normal;
    
    std::map<DegreeOfFreedom, bool> visited;

    const BasisFunction *basFcts = feSpace->getBasisFcts();
    int nBasisFcts = basFcts->getNumber();
    
    TraverseStack stack;
    ElInfo *elInfo = stack.traverseFirst(mesh, -1,
      Mesh::CALL_LEAF_EL | Mesh::FILL_BOUND | Mesh::FILL_COORDS);

    std::vector<DegreeOfFreedom> localIndices(nBasisFcts);
	
	// for pressure
    const FiniteElemSpace* feSpaceP = getFeSpace(dow);
    const BasisFunction *basFctsP = feSpaceP->getBasisFcts();
    int nBasisFctsP = basFctsP->getNumber();
    mtl::dense_vector<double> pLocalCoeffs(nBasisFctsP);

    while (elInfo) {
      for (int face = 0; face < dim + 1; face++) {
        if (elInfo->getBoundary(face) != 0) { //boundaryType) {
		  elInfo->getNormal(face, normal);
		  basFcts->getLocalIndices(elInfo->getElement(), feSpace->getAdmin(), localIndices);
		  
		  // pressure
		  getSolution()->getDOFVector(dow)->getLocalVector(elInfo->getElement(), pLocalCoeffs);

		  for (int i = 0; i < nBasisFcts; i++) {
			if (!visited[localIndices[i]]) {				
				
			  double p = basFctsP->evalUh(*basFcts->getCoords(i), pLocalCoeffs);

			  for (int j = 0; j < Global::getGeo(WORLD); j++) {
				double tmp = NormalStress(j, density*viscosity)(normal, (*grdU[0])[localIndices[i]], (*grdU[1])[localIndices[i]]);
				double result = NormalStressP(j)(normal, tmp, p);
				(*stress)[localIndices[i]][j] = result;
			  }
			  visited[localIndices[i]] = true;
			}
		  }
        }
      }

      elInfo = stack.traverseNext(elInfo);
    }
    
	for (size_t i = 0; i < dow; i++)
	  delete grdU[i];
  }
  
//   Polygon* obstacle;
  InflowBC* inflowBC;
  FileVectorWriter *fileWriterStress;
  
};


int main(int argc, char** argv)
{ FUNCNAME("main");

  AMDiS::init(argc, argv);

  NS_Channel nsProb("ns");
  Elasticity_Obstacle elastProb("elasticity->obstacle");
  Elasticity_Fluid elastProb2("elasticity->fluid");

  ElasticityNavierStokes<Elasticity_Obstacle, Elasticity_Fluid, NS_Channel> mainProb("main", &elastProb, &elastProb2, &nsProb);

  // Adapt-Infos
  AdaptInfo adaptInfo("adapt", 7);
  mainProb.initialize(&adaptInfo);

  // adaption loop - solve ch-prob and ns-prob
  AdaptInstationary adaptInstat("adapt", mainProb, adaptInfo, mainProb, adaptInfo);

  // scale Mesh
  double H = 4.1, L = 25.0;
  Parameters::get("mesh->H",H);
  Parameters::get("mesh->L",L);
//   WorldVector<double> scale; scale[0] = L; scale[1] = H;
//   Helpers::scaleMesh(nsProb.getMesh(), scale);

  ptime start_time = microsec_clock::local_time();
  int error_code = adaptInstat.adapt(); 
  time_duration td = microsec_clock::local_time()-start_time;

  MSG("elapsed time= %d sec\n", td.total_seconds());

  AMDiS::finalize();

  return error_code;
};
