/******************************************************************************
 *
 * Extension of AMDiS - Adaptive multidimensional simulations
 *
 * Copyright (C) 2013 Dresden University of Technology. All Rights Reserved.
 * Web: https://fusionforge.zih.tu-dresden.de/projects/amdis
 *
 * Authors: Simon Praetorius et al.
 *
 * This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
 * WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 *
 * See also license.opensource.txt in the distribution.
 * 
 ******************************************************************************/

namespace AMDiS { namespace extensions { 
  
void MTLPreconNavierStokes::init(const BlockMatrix& A_, const MatrixType& fullMatrix_)
{
  super::init(A_, fullMatrix_);
  
  // === extract velocity rows === 
  BlockMapper mapper(A_);
  mapper.setRow(Global::getGeo(WORLD));
  vel_rows.set(0, mapper.row(0));
  
  // === create submatrices === 
  Matrix<DOFMatrix*>* mat = A_.getOriginalMat();
  for (int i = 0; i < Global::getGeo(WORLD); i++) {
    for (int j = 0; j < Global::getGeo(WORLD); j++) {
      velocityMatrix_[i][j] = (*mat)[i][j];
    }
  }
  blockVelocityMatrix.setMatrix(velocityMatrix_);
  runner_velocity.init(blockVelocityMatrix, getMatV()); // TODO: eventuell auf neues Erzeugen verzichten!
  
  for (int j = 0; j < Global::getGeo(WORLD); j++)
    divMatrix_[0][j] = (*mat)[pressureComponent][j];
  blockDivMatrix.setMatrix(divMatrix_);
  
  for (int i = 0; i < Global::getGeo(WORLD); i++)
    gradMatrix_[i][0] = (*mat)[i][pressureComponent];
  blockGradMatrix.setMatrix(gradMatrix_);

  
  const FiniteElemSpace *pressureFeSpace = componentSpaces[pressureComponent];
  
  // === Mass matrix solver ===    
  massMatrix = new DOFMatrix(pressureFeSpace, pressureFeSpace);
  {
    Operator massOp(pressureFeSpace, pressureFeSpace);
    if ((!phase) || (*nu == 0.0))
      addZOT(massOp, 1.0);
    else
      addZOT(massOp, valueOf(phase));
    massMatrix->assembleOperator(massOp);
  }
  massMatrix_(0,0) = massMatrix;
  blockMassMatrix.setMatrix(massMatrix_);
  runner_mass.init(blockMassMatrix, getM());
  

  // === Laplace matrix solver ===
  laplaceMatrix = new DOFMatrix(pressureFeSpace, pressureFeSpace);
  {
    Operator laplaceOp(pressureFeSpace, pressureFeSpace);
    if ((!phase) || (*nu == 0.0))
      addSOT(laplaceOp, 1.0);
    else
      addSOT(laplaceOp, valueOf(phase));
    laplaceMatrix.assembleOperator(laplaceOp);
  }
  laplaceMatrix_(0,0) = laplaceMatrix;
  blockLaplaceMatrix.setMatrix(laplaceMatrix_);
  runner_laplace.init(blockLaplaceMatrix, getL());
  
  
  // === Create convection-diffusion operator ===
  conDifMatrix = new DOFMatrix(pressureFeSpace, pressureFeSpace);
  {
    Operator conDifOp(pressureFeSpace, pressureFeSpace);
    
    DOFVector<double> vp(pressureFeSpace, "vp");
    
    WorldVector<DOFVector<double>*> V;
    for (int i = 0; i < V.getSize(); i++) {
      V[i] = new DOFVector<double>(pressureFeSpace, "v_i");
      V[i]->interpol(solution->getDOFVector(i));
    }

    if (!phase) { // no phase given
      MSG("INIT WITHOUT PHASE!\n");
      addZOT(conDifOp, *invTau);
      addZOT(conDifOp, *nu);
      addFOT(conDifOp, valueOf(V), GRD_PHI);
    } else {
      vp.interpol(phase);
      
      if (*nu > 0.0) {
	addZOT(conDifOp, valueOf(vp) * (*invTau));
	addSOT(conDifOp, valueOf(vp) * (*nu));
	addFOT(conDifOp, valueOf(vp) * valueOf(V), GRD_PHI);
      } else {
	addZOT(conDifOp, function_(wrap(new LinearInterpolation(*rho1,*rho2,*invTau)), valueOf(vp)) );
	addSOT(conDifOp, function_(wrap(new LinearInterpolation(*nu1,*nu2)), valueOf(vp)) );
	addFOT(conDifOp, valueOf(V) * function_(wrap(new LinearInterpolation(*rho1,*rho2)), valueOf(vp)), GRD_PHI );
      }
    }
    conDifMatrix->assembleOperator(conDifOp);
    
    for (int i = 0; i < V.getSize(); i++)
      delete V[i];
  }

  // temporary variables
  y01.change_dim(num_rows(getMatV()));
  y2.change_dim(num_rows(getM()));
}


void MTLPreconNavierStokes::solve(const VectorType& b, VectorType& x) const
{ FUNCNAME("MTLPreconNavierStokes::solve()");

  x.change_dim(num_rows(b));
  
  const VectorType b01(b[vel_rows]);
  const VectorType b2(b[rows[pressureComponent]]);
    
  VectorType x01(x[vel_rows]);
  VectorType x2(x[rows[pressureComponent]]);
  
  runner_velocity->solve(getMatV(), y01, b01);	// -> y01
  y2 = b2 - getMatDiv() * y01;			// -> y2
  runner_laplace->solve(getL(), x2, y2);		// -> x2
  
  // project out constant Null-space
  double vecSum = sum(x2) / size(x2);
  x2 -= vecSum;					// -> x2 -= sum(x2)/size(x2)
    
  y2 = getMatConDif() * x2;			// -> y2
  runner_mass->solve(getM(), x2, y2);		// => x2
  
  y01 = b01 - getMatGrad() * x2;		// -> y01
  runner_velocity->solve(getMatV(), x01, y01);	// => x01
}

} }
