/******************************************************************************
 *
 * Extension of AMDiS - Adaptive multidimensional simulations
 *
 * Copyright (C) 2013 Dresden University of Technology. All Rights Reserved.
 * Web: https://fusionforge.zih.tu-dresden.de/projects/amdis
 *
 * Authors: Simon Praetorius et al.
 *
 * This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
 * WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 *
 * See also license.opensource.txt in the distribution.
 * 
 ******************************************************************************/

namespace AMDiS { namespace extensions { 
  
template<typename MatrixType>
void MTLPreconPfcBase<MatrixType>::init(const SolverMatrix<Matrix<DOFMatrix*> >& A_, const MatrixType& fullMatrix_)
{
  assert(tau != NULL);  
  super::init(A_, fullMatrix_);
  
  typedef typename mtl::Collection<MTLMatrix>::size_type size_type;
  typedef typename mtl::Collection<MTLMatrix>::value_type value_type;
  size_type n = num_rows(getM());
  
  double delta = std::sqrt(M0 * getTau());
  
  // helper-matrix MpL = M + delta*L
  MpL.change_dim(n, n);
  MpL = getM() + (1.0/delta)*getL0();

  // helper-matrix MpL = M + sqrt(delta)*L
  MpL2.change_dim(n, n);
  MpL2 = getM() + ( std::sqrt(delta) ) * getL();
  
  
  bool useDirectSolverM = false, 
       useDirectSolverMpL = false, 
       useDirectSolverMpL2 = false;
  Parameters::get("precon_pfc_M->use direct solver", useDirectSolverM);
  Parameters::get("precon_pfc_MpL->use direct solver", useDirectSolverMpL);
  Parameters::get("precon_pfc_MpL2->use direct solver", useDirectSolverMpL2);
  
  PId = new itl::pc::identity<MTLTypes::MTLMatrix, double>(getM());
  
  if (useDirectSolverM)
    solverM = new mtl::matrix::umfpack::solver<MTLTypes::MTLMatrix>(getM());
  else
    PDiagM = new itl::pc::diagonal<MTLTypes::MTLMatrix, double>(getM());
  if (useDirectSolverMpL)
    solverMpL = new mtl::matrix::umfpack::solver<MTLTypes::MTLMatrix>(MpL);
  else
    PDiagMpL = new itl::pc::diagonal<MTLTypes::MTLMatrix, double>(MpL);
  if (useDirectSolverMpL2)
    solverMpL2 = new mtl::matrix::umfpack::solver<MTLTypes::MTLMatrix>(MpL2);
  else
    PDiagMpL2 = new itl::pc::diagonal<MTLTypes::MTLMatrix, double>(MpL2);
    
  // temporary variables
  y0.change_dim(num_rows(getM()));
  y1.change_dim(num_rows(getM()));
  tmp.change_dim(num_rows(getM()));
}


template<typename MatrixType>
void MTLPreconPfcBase<MatrixType>::solve(const MTLVector& b, MTLVector& x) const
{ FUNCNAME("MTLPreconPfc::solve()");

  x.change_dim(num_rows(b));
  
  MTLVector x0(x[super::getRowRange(0)]);
  MTLVector x1(x[super::getRowRange(1)]);
  MTLVector x2(x[super::getRowRange(2)]);
  
  const MTLVector b0(b[super::getRowRange(0)]);
  const MTLVector b1(b[super::getRowRange(1)]);
  const MTLVector b2(b[super::getRowRange(2)]); 
  
  double delta = std::sqrt(M0 * getTau());
  
  solveM(y0, b0);		// M*y0 = b0
  y1 = getL0() * y0;		// y1 = K*y0
  tmp = b1 - y1;		// tmp := b1 - tau*y1

  solveMpL(y1, tmp);		// (M + delta*K) * y1 = tmp
  x0 = y0 + (1.0/delta)*y1;	// x0 = y0 + (1/delta)*y1

  tmp = getM() * y1;		// tmp := M*y1
  solveMpL2(y1, tmp);		// (M+eps*sqrt(delta)K) * y1 = tmp
  tmp = getM() * y1;		// tmp := M*y1
  solveMpL2(x1, tmp);		// (M+eps*sqrt(delta)K) * x1 = tmp

  x0-= (1.0/delta)*x1;		// x0 = x0 - (1/delta)*x1 = y0 + (1/delta)*(y1 - x1)

  tmp = b2 - getL() * x1;	// tmp := b2 - K*x1
  solveM(x2, tmp);
}

} }
