/******************************************************************************
 *
 * Extension of AMDiS - Adaptive multidimensional simulations
 *
 * Copyright (C) 2013 Dresden University of Technology. All Rights Reserved.
 * Web: https://fusionforge.zih.tu-dresden.de/projects/amdis
 *
 * Authors: Simon Praetorius et al.
 *
 * This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
 * WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 *
 * See also license.opensource.txt in the distribution.
 * 
 ******************************************************************************/
#ifndef MTL_PFC_PRECONDITIONER_DIAG_H
#define MTL_PFC_PRECONDITIONER_DIAG_H

#include "AMDiS.h"
#include "BlockPreconditioner.h"

namespace AMDiS { namespace extensions { 

using namespace std; 
using namespace AMDiS::MTLTypes;

struct MTLPreconPfcDiag : BlockPreconditioner
{
  typedef BlockPreconditioner super;
 
  class Creator : public PreconditionCreator
  {
  public:
    virtual ~Creator() {}
    
    BasePreconditioner* create() { 
      return new MTLPreconPfcDiag(this->name);
    }
  };  
  
  MTLPreconPfcDiag(std::string name)
  : BlockPreconditioner(),
    name(name), 
    tau(NULL),
    PId(NULL), PDiagM(NULL), PDiagMpL(NULL), PDiagS(NULL),
    solverM(NULL), solverMpL(NULL), solverS(NULL)
  { }
  
  virtual ~MTLPreconPfcDiag()
  {
    exit();
  }
  
  void setData(double* tau_)
  {
    tau = tau_;
  }
  
  virtual void init(const typename super::BlockMatrix& A, const MTLTypes::MTLMatrix& fullMatrix);
  
  virtual void exit()
  {    
    if (PId) { delete PId; PId = NULL; }
    if (PDiagM) { delete PDiagM; PDiagM = NULL; }
    if (PDiagMpL) { delete PDiagMpL; PDiagMpL = NULL; }
    if (PDiagS) { delete PDiagS; PDiagS = NULL; }
    if (solverM) { delete solverM; solverM = NULL; }
    if (solverMpL) { delete solverMpL; solverMpL = NULL; }
    if (solverS) { delete solverS; solverS = NULL; }
  }
  
  virtual void solve(const MTLVector& b, MTLVector& x) const;  
    
  template <typename VectorX, typename VectorB>
  void solveM(VectorX& x, const VectorB& b, int nIter = -1) const
  {
    int nIterM=5;
    double rtolM=0.0, tolM=0.0;
    bool useDirectSolver = false;
    Parameters::get("precon_pfc_M->max iteration", nIterM);
    Parameters::get("precon_pfc_M->tolerance", tolM);
    Parameters::get("precon_pfc_M->relative tolerance", rtolM);
    Parameters::get("precon_pfc_M->use direct solver", useDirectSolver);
    if (useDirectSolver && solverM)
      (*solverM)(x,b);
    else {
      itl::basic_iteration<double> iter(b, nIterM, rtolM, tolM);
      x = 0.0;
      itl::cg(getM(), x, b, *PDiagM, *PId, iter);
    }
  }
  
  template <typename VectorX, typename VectorB>
  void solveMpL(VectorX& x, const VectorB& b) const
  {
    int nIterMpL=20;
    double rtolMpL=0.0, tolMpL=0.0;
    bool useDirectSolver = false;
    Parameters::get("precon_pfc_MpL->max iteration", nIterMpL);
    Parameters::get("precon_pfc_MpL->tolerance", tolMpL);
    Parameters::get("precon_pfc_MpL->relative tolerance", rtolMpL);
    Parameters::get("precon_pfc_MpL->use direct solver", useDirectSolver);
    if (useDirectSolver && solverMpL)
      (*solverMpL)(x,b);
    else {
      itl::basic_iteration<double> iter(b, nIterMpL, rtolMpL, tolMpL);
      x = 0.0;
      itl::cg(MpL, x, b, *PDiagMpL, *PId, iter);
    }
  }
  
  template <typename VectorX, typename VectorB>
  void solveS(VectorX& x, const VectorB& b) const
  {
    int nIterMpL2=10;
    double rtolMpL2=0.0, tolMpL2=0.0;
    bool useDirectSolver = false;
    Parameters::get("precon_pfc_MpL2->max iteration", nIterMpL2);
    Parameters::get("precon_pfc_MpL2->tolerance", tolMpL2);
    Parameters::get("precon_pfc_MpL2->relative tolerance", rtolMpL2);
    Parameters::get("precon_pfc_MpL2->use direct solver", useDirectSolver);
    if (useDirectSolver && solverS)
      (*solverS)(x,b);
    else {
      itl::basic_iteration<double> iter(b, nIterMpL2, rtolMpL2, tolMpL2);
      x = 0.0;
      
      std::string solver = "cg";
      Parameters::get("precon_pfc_MpL2->solver", solver);
      if (solver == "cg")
	itl::cg(S, x, b, *PDiagS, *PId, iter);
      else if (solver == "gmres")
	itl::gmres(S, x, b, *PDiagS, *PId, iter, 30);
      else if (solver == "tfqmr")
	itl::tfqmr(S, x, b, *PDiagS, *PId, iter);
      else if (solver == "bicgstab")
	itl::bicgstab(S, x, b, *PDiagS, iter);
      else
	throw std::runtime_error("unknown sub-solver");
    }
  }
  
  
  const MTLMatrix& getM() const { return A->getSubMatrix(2,2); }
  const MTLMatrix& getL() const { return A->getSubMatrix(2,1); }
  double getTau() const { return *tau; }
  
protected: 
  MTLMatrix MpL;
  MTLMatrix MinvL;
  MTLMatrix S;
   
  mutable MTLVector y0;
  mutable MTLVector y1;
  mutable MTLVector tmp;
  
  itl::pc::identity<MTLTypes::MTLMatrix, double> *PId;
  itl::pc::diagonal<MTLTypes::MTLMatrix, double> *PDiagM;
  itl::pc::diagonal<MTLTypes::MTLMatrix, double> *PDiagMpL;
  itl::pc::diagonal<MTLTypes::MTLMatrix, double> *PDiagS;
  
  mtl::matrix::umfpack::solver<MTLTypes::MTLMatrix> *solverM;
  mtl::matrix::umfpack::solver<MTLTypes::MTLMatrix> *solverMpL;
  mtl::matrix::umfpack::solver<MTLTypes::MTLMatrix> *solverS;
  
  std::string name;
  
  double* tau;
};

} }

#include "MTLPreconPfc_diag.hh"

#endif // MTL_PFC_PRECONDITIONER_H

